﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Net;
using System.IO;
using System.Configuration;
using System.Windows.Forms;

using SkyLordGateClient.UserInterface;
using SkyLordGateClient.BusinessLogic.TextChat;
using SkyLordGateClient.BusinessLogic.VideoChat;
using SkyLordGateClient.BusinessLogic.Database;

namespace SkyLordGateClient.BusinessLogic
{
    public sealed class BusinessLogicManager 
    {
        private static BusinessLogicManager instance = new BusinessLogicManager(); // static private object to make it singleton
        private int userID;        // user ID of the currently logged on user. to be set after user authorization.
        private string selectedTabName; //Name of the tab selected by the user.

        private TextChatHelper textChatHelper;
        private VideoChatHelper videoChatHelper;
        
        private List<VideoChatControl> vList;
        private StreamWriter logFileWriter;
        
        /// <summary>
        /// Constructor for BusinessLogicManager Class
        /// </summary>
        private BusinessLogicManager() 
        {          
            userID = -1;
            selectedTabName = "";
            logFileWriter = File.AppendText(ConfigurationManager.AppSettings["LogFilePath"]);
            
            textChatHelper = new TextChatHelper();            
        }

        /// <summary>
        /// 
        /// </summary>
        public void InitVideoService()
        {
            videoChatHelper = new VideoChatHelper(DatabaseHelper.Instance.GetUserIPAddress(userID), DatabaseHelper.Instance.GetUserVideoPort(userID));
        }

        /// <summary>
        /// Singleton implementation of Business Manager.
        /// </summary>
        public static BusinessLogicManager Instance 
        {
            get
            {
                return instance;
            }
        }

        /// <summary>
        /// Return userid of the logged on user.
        /// </summary>
        /// <returns></returns>
        public int GetUserID() 
        {
            return userID;
        }

        /// <summary>
        /// Sets the userid of the logged on user.
        /// </summary>
        /// <param name="uID"></param>
        public void SetUserID(int uID) 
        {
            userID = uID;
        }

        /// <summary>
        /// Returns the user nickname from database for the given uID.
        /// </summary>
        /// <returns></returns>
        public string GetUserName(int uID) 
        {
            return DatabaseHelper.Instance.GetUserName(uID);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="uID"></param>
        /// <returns></returns>
        public string GetUserIPAddress(int uID) 
        {
            return DatabaseHelper.Instance.GetUserIPAddress(uID);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="uID"></param>
        /// <returns></returns>
        public int GetUserVideoPort(int uID) 
        {
            return DatabaseHelper.Instance.GetUserVideoPort(uID);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="tabName"></param>
        public void SetSelectedTabName(string tabName) 
        {
            selectedTabName = tabName;
        }

        public string GetSelectedTabName() 
        {
            return selectedTabName;
        }

        /// <summary>
        /// Returns the list of video devices from video capture library.
        /// </summary>
        /// <returns></returns>
        public List<string> GetVideoDevices() 
        {
            return videoChatHelper.GetVideoDevices();
        }

        /// <summary>
        /// Sets the video chat control list
        /// </summary>
        /// <param name="videoList"></param>
        public void SetVideoChatControlList(ref List<VideoChatControl> videoList)
        {
            vList = videoList;
        }

        /// <summary>
        /// Stop the video streaming service and other resources
        /// </summary>
        public void OnExit() 
        {
            // stop video service
            videoChatHelper.OnStop();

            // disconnect user from this tab
            textChatHelper.LeaveTab(userID, selectedTabName);

            // close the log file stream
            logFileWriter.Close();
        }

        /// <summary>
        /// Returns tab names for the logged in user
        /// </summary>
        /// <returns></returns>
        public List<string> GetTabNames()
        {
            return DatabaseHelper.Instance.GetTabNames(userID);           
        }

        /// <summary>
        /// Returns User IDs list for this tab
        /// </summary>
        /// <param name="tabIndexOnGUI"></param>
        /// <returns></returns>
        public Hashtable GetUsersForTab(string tabName) 
        {
            Hashtable usersOnline = new Hashtable();

            List<int> connectedUsersOnTab = DatabaseHelper.Instance.GetConnectedUsersOnTab(userID, tabName);
            List<int> allUsersOnTab = DatabaseHelper.Instance.GetUsersForTab(userID, tabName);
            foreach (int user in allUsersOnTab) 
            {
                if (connectedUsersOnTab.Contains(user))
                    usersOnline.Add(user, true);
                else
                    usersOnline.Add(user, false);
            }
            return usersOnline;
        }

        /// <summary>
        /// Makes a connection to the server that specifies that the logged in user has connected 
        /// to tabname provided in parameters.
        /// </summary>
        /// <param name="tabName">Name of the tab to connect to.</param>
        public void ConnectToTab()
        {
            textChatHelper.ConnectToTab(userID, selectedTabName);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="newTabName"></param>
        public void SwitchTabs(string newTabName)
        {
            textChatHelper.SwitchTabs(userID, selectedTabName, newTabName);
            selectedTabName = newTabName;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="receiverID"></param>
        /// <param name="message"></param>
        public void SendTextMessage(int receiverID, string message) 
        {
            int tabID = DatabaseHelper.Instance.GetTabID(userID, selectedTabName);
            textChatHelper.SendTextMessage(userID, tabID, receiverID, message);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="senderID"></param>
        /// <param name="tabID"></param>
        /// <param name="message"></param>
        public void RecieveTextMessage(int senderID, int tabID, string message) 
        {
            int selectedTabID = DatabaseHelper.Instance.GetTabID(userID, selectedTabName);
            if (tabID == selectedTabID)
            {
                for (int i = 0; i < vList.Count; i++)
                    if (vList[i].GetUserID() == senderID)
                        vList[i].ReceiveTextMessage(message);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="message"></param>
        public void WriteLogMessage(string message) 
        {
            lock (logFileWriter)
            {
                try
                {
                    logFileWriter.WriteLine(DateTime.Now.ToString() + " " + message);
                    logFileWriter.Flush();
                }
                catch { }
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="userID"></param>
        /// <param name="tabID"></param>
        public void ChangeUserStatus(int uID, int tabID, bool isOnline)
        {
            if (uID == userID) 
            {
                MessageBox.Show("Connection to server lost. Closing application");
                WriteLogMessage("Connection to server lost. Closing application");
                Environment.Exit(1);
            }

            int selectedTabID = DatabaseHelper.Instance.GetTabID(userID, selectedTabName);
            if (tabID == selectedTabID)
            {
                for (int i = 0; i < vList.Count; i++)
                    if (vList[i].GetUserID() == uID)
                        vList[i].SetStatus(isOnline);
            }
        }
    }
}

